@@ -627,7 +627,10 @@ process_numeric (session * sess, int n,
 	case 349:	/* end of exemption list */
 		sess = find_channel (serv, word[4]);
 		if (!sess)
+		{
 			sess = serv->front_session;
+			goto def;
+		}
 		if (!fe_is_banwindow (sess))
 			goto def;
 		fe_ban_list_end (sess, TRUE);
@@ -650,12 +653,21 @@ process_numeric (session * sess, int n,
 	case 368:
 		sess = find_channel (serv, word[4]);
 		if (!sess)
+		{
 			sess = serv->front_session;
+			goto def;
+		}
 		if (!fe_is_banwindow (sess))
 			goto def;
 		fe_ban_list_end (sess, FALSE);
 		break;
 
+	case 369:	/* WHOWAS end */
+	case 406:	/* WHOWAS error */
+		EMIT_SIGNAL (XP_TE_SERVTEXT, whois_sess, text, word[1], NULL, NULL, 0);
+		serv->inside_whois = 0;
+		break;
+
 	case 372:	/* motd text */
 	case 375:	/* motd start */
 		if (!prefs.skipmotd || serv->motd_skipped)
@@ -668,7 +680,8 @@ process_numeric (session * sess, int n,
 		inbound_login_end (sess, text);
 		break;
 
-	case 433:
+	case 433:	/* nickname in use */
+	case 432:	/* erroneous nickname */
 		if (serv->end_of_motd)
 			goto def;
 		inbound_next_nick (sess,  word[4]);
