@@ -18,9 +18,10 @@ void mg_set_access_icon (session_gui *gui, GdkPixbuf *pix, gboolean away);
 void mg_apply_setup (void);
 void mg_close_sess (session *);
 void mg_tab_close (session *sess);
-void mg_link_cb (GtkWidget *but, gpointer userdata);
+void mg_detach (session *sess, int mode);
 void mg_progressbar_create (session_gui *gui);
 void mg_progressbar_destroy (session_gui *gui);
 void mg_dnd_drop_file (session *sess, char *target, char *uri);
 void mg_change_layout (int type);
 void mg_update_meters (session_gui *);
+void mg_inputbox_cb (GtkWidget *igad, session_gui *gui);
